@@ -2,37 +2,42 @@
 
 # spechomo, Spectral homogenization of multispectral satellite data
 #
-# Copyright (C) 2020  Daniel Scheffler (GFZ Potsdam, daniel.scheffler@gfz-potsdam.de)
+# Copyright (C) 2019-2021
+# - Daniel Scheffler (GFZ Potsdam, daniel.scheffler@gfz-potsdam.de)
+# - Helmholtz Centre Potsdam - GFZ German Research Centre for Geosciences Potsdam,
+#   Germany (https://www.gfz-potsdam.de/)
 #
 # This software was developed within the context of the GeoMultiSens project funded
 # by the German Federal Ministry of Education and Research
 # (project grant code: 01 IS 14 010 A-C).
 #
-# This program is free software: you can redistribute it and/or modify it under
-# the terms of the GNU Lesser General Public License as published by the Free
-# Software Foundation, either version 3 of the License, or (at your option) any
-# later version. Please note the following exception: `spechomo` depends on tqdm,
-# which is distributed under the Mozilla Public Licence (MPL) v2.0 except for the
-# files "tqdm/_tqdm.py", "setup.py", "README.rst", "MANIFEST.in" and ".gitignore".
+# Licensed under the Apache License, Version 2.0 (the "License");
+# you may not use this file except in compliance with the License.
+# You may obtain a copy of the License at
+#
+#   http://www.apache.org/licenses/LICENSE-2.0
+#
+# Please note the following exception: `spechomo` depends on tqdm, which is
+# distributed under the Mozilla Public Licence (MPL) v2.0 except for the files
+# "tqdm/_tqdm.py", "setup.py", "README.rst", "MANIFEST.in" and ".gitignore".
 # Details can be found here: https://github.com/tqdm/tqdm/blob/master/LICENCE.
 #
-# This program is distributed in the hope that it will be useful, but WITHOUT
-# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
-# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
-# details.
-#
-# You should have received a copy of the GNU Lesser General Public License along
-# with this program.  If not, see <http://www.gnu.org/licenses/>.
+# Unless required by applicable law or agreed to in writing, software
+# distributed under the License is distributed on an "AS IS" BASIS,
+# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+# See the License for the specific language governing permissions and
+# limitations under the License.
 
 """Main module."""
 
 import os
-import numpy as np
 import logging  # noqa F401  # flake8 issue
 from typing import Union, Tuple  # noqa F401  # flake8 issue
 from multiprocessing import cpu_count
 import traceback
 import time
+from tqdm import tqdm
+import numpy as np
 from geoarray import GeoArray  # noqa F401  # flake8 issue
 from specclassify import classify_image
 # from specclassify import kNN_MinimumDistance_Classifier
@@ -41,6 +46,7 @@ from .classifier import Cluster_Learner
 from .exceptions import ClassifierNotAvailableError
 from .logging import SpecHomo_Logger
 from .options import options
+from .utils import spectra2im, im2spectra
 
 
 __author__ = 'Daniel Scheffler'
@@ -51,15 +57,17 @@ _classifier_rootdir = options['classifiers']['rootdir']
 class SpectralHomogenizer(object):
     """Class for applying spectral homogenization by applying an interpolation or machine learning approach."""
 
-    def __init__(self, classifier_rootDir='', logger=None, CPUs=None):
+    def __init__(self, classifier_rootDir='', logger=None, CPUs=None, progress=True):
         """Get instance of SpectralHomogenizer.
 
         :param classifier_rootDir:  root directory where machine learning classifiers are stored.
         :param logger:              instance of logging.Logger
+        :param progress:            whether to show progress bars
         """
         self.classifier_rootDir = classifier_rootDir or _classifier_rootdir
         self.logger = logger or SpecHomo_Logger(__name__)
         self.CPUs = CPUs or cpu_count()
+        self.progress = progress
 
     def interpolate_cube(self, arrcube, source_CWLs, target_CWLs, kind='linear'):
         # type: (Union[np.ndarray, GeoArray], list, list, str) -> GeoArray
@@ -155,7 +163,7 @@ class SpectralHomogenizer(object):
                                       - only usable for 'MinDist', 'SAM' and 'SID' as well as their kNN variants
                                       - may be given as float, integer or string to label a certain distance percentile
                                       - if given as string, it must match the format, e.g., '10%' for labelling the
-                                      worst 10 % of the distances as unclassified
+                                        worst 10 % of the distances as unclassified
         :param src_nodataVal:       no data value of source image (arrcube)
                                     - if no nodata value is set, it is tried to be auto-computed from arrcube
         :param out_nodataVal:       no data value of predicted image
@@ -174,6 +182,7 @@ class SpectralHomogenizer(object):
                   n_clusters=n_clusters,
                   classif_alg=classif_alg,
                   CPUs=self.CPUs,
+                  progress=self.progress,
                   logger=self.logger)
 
         if classif_alg.startswith('kNN'):
@@ -257,8 +266,8 @@ class RSImage_ClusterPredictor(object):
     """Predictor class applying the predict() function of a machine learning classifier described by the given args."""
 
     def __init__(self, method='LR', n_clusters=50, classif_alg='MinDist', classifier_rootDir='',
-                 CPUs=1, logger=None, **kw_clf_init):
-        # type: (str, int, str, str, Union[None, int], logging.Logger, dict) -> None
+                 CPUs=1, logger=None, progress=True, **kw_clf_init):
+        # type: (str, int, str, str, Union[None, int], logging.Logger, bool, dict) -> None
         """Get an instance of RSImage_ClusterPredictor.
 
         :param method:              machine learning approach to be used for spectral bands prediction
@@ -284,6 +293,7 @@ class RSImage_ClusterPredictor(object):
                                     'kNN_FEDSA': k-nearest-neighbour fused euclidian distance / spectral angle
         :param classifier_rootDir:  root directory where machine learning classifiers are stored.
         :param CPUs:                number of CPUs to use (default: 1)
+        :param progress:            whether to show progress bars
         :param logger:              instance of logging.Logger()
         :param kw_clf_init          keyword arguments to be passed to classifier init functions if possible,
                                     e.g., 'n_neighbours' sets the number of neighbours to be considered in kNN
@@ -298,6 +308,7 @@ class RSImage_ClusterPredictor(object):
         self.CPUs = CPUs or cpu_count()
         self.classif_alg = classif_alg
         self.logger = logger or SpecHomo_Logger(__name__)  # must be pickable
+        self.progress = progress
         self.kw_clf_init = kw_clf_init
 
         # validate
@@ -362,8 +373,6 @@ class RSImage_ClusterPredictor(object):
         # type: (Union[np.ndarray, GeoArray], Cluster_Learner, float, float, float, Union[str, int, float], int) -> GeoArray  # noqa
         """Apply the prediction function of the given specifier to the given remote sensing image.
 
-        # NOTE: The 'nodataVal' is written
-
         :param image:           3D array representing the input image
         :param classifier:      the classifier instance
         :param in_nodataVal:    no data value of the input image
@@ -396,6 +405,8 @@ class RSImage_ClusterPredictor(object):
         # assign each input pixel to a cluster (compute classification with cluster centers as endmembers)
         if self.classif_map is None:
             if self.n_clusters > 1:
+                self.logger.info('Assigning material-specific regressors to each image pixel.')
+
                 t0 = time.time()
                 kw_clf = dict(classif_alg=self.classif_alg,
                               in_nodataVal=image.nodata,
@@ -417,6 +428,8 @@ class RSImage_ClusterPredictor(object):
                     train_spectra = classifier.cluster_centers
                     train_labels = classifier.cluster_pixVals
 
+                # run classification
+                # - uses 3 neighbors by default in case of kNN classifiers
                 self.classif_map, self.distance_metrics = classify_image(image, train_spectra, train_labels, **kw_clf)
 
                 # compute spectral distance
@@ -445,33 +458,16 @@ class RSImage_ClusterPredictor(object):
                 self.distance_metrics = np.zeros_like(self.classif_map,
                                                       np.float32)
 
-        ####################
-        # apply prediction #
-        ####################
-
-        # adjust classifier
-        if self.CPUs is None or self.CPUs > 1:
-            # FIXME does not work -> parallelize with https://github.com/ajtulloch/sklearn-compiledtrees?
-            classifier.n_jobs = cpu_count() if self.CPUs is None else self.CPUs
-
-        # NOTE: prediction is applied in 1000 x 1000 tiles to save memory (because classifier.predict returns float32)
-        t0 = time.time()
-        out_nodataVal = out_nodataVal if out_nodataVal is not None else image.nodata
-        image_predicted = GeoArray(np.empty((image.rows,
-                                             image.cols,
-                                             classifier.tgt_n_bands),
-                                            dtype=image.dtype),
-                                   geotransform=image.gt,
-                                   projection=image.prj,
-                                   nodata=out_nodataVal,
-                                   bandnames=['B%s' % i
-                                              if len(i) == 2
-                                              else 'B0%s' % i
-                                              for i in classifier.tgt_LBA])
+        ##############################
+        # compute prediction weights #
+        ##############################
 
+        # compute the weights (only needed in case of multiple kNN classifiers)
         if classifier.n_clusters > 1 and\
            self.classif_map.ndim > 2:
 
+            self.logger.info('Computing prediction weights per pixel for each regressor.')
+
             if self.classif_alg == 'kNN_SAM':
                 # scale SAM values between 0 and 15 degrees spectral angle
                 dist_min, dist_max = 0, 15
@@ -487,7 +483,7 @@ class RSImage_ClusterPredictor(object):
             dist_norm = (self.distance_metrics - dist_min) /\
                         (dist_max - dist_min)
             weights = 1 - dist_norm
-            weights[weights < 0] = 1e-10  # FIXME 0 causes ZeroDivisionError later
+            weights[weights < 0] = 1e-10  # set negative weights to 0 but avoid ZeroDivisionError
 
         else:
             weights = None
@@ -499,30 +495,92 @@ class RSImage_ClusterPredictor(object):
         #     print(self.distance_metrics[0, 0, :])
         #     print(weights[0, 0, :])
 
-        n_saturated_px = 0
-        for ((rS, rE), (cS, cE)), im_tile in image.tiles(tilesize=(1000, 1000)):
-            self.logger.info('Predicting tile ((%s, %s), (%s, %s))...' % (rS, rE, cS, cE))
+        ####################
+        # apply prediction #
+        ####################
 
-            classif_map_tile = self.classif_map[rS: rE + 1, cS: cE + 1]  # integer array
+        self.logger.info(f'Starting prediction with {self.method} regressor, {self.n_clusters} clusters, '
+                         f'{self.classif_alg}.')
+
+        # adjust classifier for multiprocessing
+        if self.CPUs is None or self.CPUs > 1:
+            # FIXME does not work -> parallelize with https://github.com/ajtulloch/sklearn-compiledtrees?
+            classifier.n_jobs = cpu_count() if self.CPUs is None else self.CPUs
+
+        # get an empty GeoArray for the prediction result
+        t0 = time.time()
+        out_nodataVal = out_nodataVal if out_nodataVal is not None else image.nodata
+        image_predicted = GeoArray(np.empty((image.rows,
+                                             image.cols,
+                                             classifier.tgt_n_bands),
+                                            dtype=image.dtype),
+                                   geotransform=image.gt,
+                                   projection=image.prj,
+                                   nodata=out_nodataVal,
+                                   bandnames=['B%s' % i
+                                              if len(i) == 2
+                                              else 'B0%s' % i
+                                              for i in classifier.tgt_LBA])
+
+        # set image_predicted to nodata at nodata positions of the input image
+        if out_nodataVal is not None:
+            image_predicted[~image.mask_nodata[:]] = out_nodataVal
+
+        # NOTE:
+        # - prediction now only runs on the remaining pixels (that contain data)
+        # - computation is running in chunks of 50,000 spectra to save memory
+        #   (classifier.predict returns float32) and speed up processing
+        # ----------------------------------------------------------------------
+
+        # get all spectra at pixels that really contain data and
+        # reshape them to represent a single image column
+        # (classifier.predict expects a 3D image-like input array)
+        spectra_at_datapos = image[image.mask_nodata[:]]
+        n_spectra = spectra_at_datapos.shape[0]
+        spectra_as_im = GeoArray(spectra2im(spectra_at_datapos, n_spectra, 1))
+
+        # get the corresponding weights and classification maps (also as one image column)
+        if weights is not None:
+            # in case of kNN classifiers
+            weights_datapos = spectra2im(weights[image.mask_nodata[:]], n_spectra, 1)
+            classif_map_datapos = spectra2im(self.classif_map[image.mask_nodata[:]], n_spectra, 1)
+        else:
+            # in case no kNN classifier was used and we don't have to respect any weights
+            weights_datapos = None
+            classif_map_datapos = self.classif_map[image.mask_nodata[:]].reshape(-1, 1)
+
+        # spectra_predicted will be filled while looping over chunks
+        spectra_predicted = np.empty((n_spectra, image_predicted.bands), image_predicted.dtype)
+        n_saturated_px = 0
+
+        for ((rS, rE), (cS, cE)), im_tile in tqdm(spectra_as_im.tiles(tilesize=(50000, 1)),
+                                                  desc='Predicting in chunks',
+                                                  disable=not self.progress):
+
+            classif_map_tile = classif_map_datapos[rS: rE + 1, cS: cE + 1]  # integer array
 
             # predict!
             if self.classif_map.ndim == 2:
                 im_tile_pred = \
-                    classifier.predict(im_tile, classif_map_tile,
+                    classifier.predict(im_tile,
+                                       classif_map_tile,
                                        nodataVal=out_nodataVal,
                                        cmap_nodataVal=cmap_nodataVal,
                                        cmap_unclassifiedVal=unclassified_pixVal)
 
             else:
-                weights_tile = weights[rS: rE + 1, cS: cE + 1]  # float array
+                weights_tile = weights_datapos[rS: rE + 1, cS: cE + 1]  # float array
 
                 im_tile_pred = \
-                    classifier.predict_weighted_averages(im_tile, classif_map_tile, weights_tile,
+                    classifier.predict_weighted_averages(im_tile,
+                                                         classif_map_tile,
+                                                         weights_tile,
                                                          nodataVal=out_nodataVal,
                                                          cmap_nodataVal=cmap_nodataVal,
                                                          cmap_unclassifiedVal=unclassified_pixVal)
 
             # set saturated pixels (exceeding the output data range with respect to the data type) to no-data
+            # NOTE: this is computed on the chunks to save memory
             if isinstance(image_predicted.dtype, np.integer):
                 out_dTMin, out_dTMax = np.iinfo(image_predicted.dtype).min,\
                                        np.iinfo(image_predicted.dtype).max
@@ -536,28 +594,15 @@ class RSImage_ClusterPredictor(object):
                     n_saturated_px += np.sum(mask_saturated)
                     im_tile_pred[mask_saturated] = out_nodataVal
 
-            image_predicted[rS:rE + 1, cS:cE + 1] = im_tile_pred.astype(image_predicted.dtype)
+            spectra_predicted[rS:rE + 1, :] = im2spectra(im_tile_pred)  # [n_spectra x n_tgt_bands]
+
+        # fill in the predicted spectra
+        image_predicted[image.mask_nodata[:]] = spectra_predicted
 
         if n_saturated_px:
             self.logger.warning("%.2f %% of the predicted pixels are saturated and set to no-data."
                                 % n_saturated_px / np.dot(*image_predicted.shape[:2]) * 100)
 
-        # TODO add multiprocessing here? ML classifiers seem to use multiprocessing already
-        # print(time.time() -t0)
-        # t0 = time.time()
-        # from multiprocessing import Pool
-        # from geoarray.baseclasses import get_array_tilebounds
-        # with Pool(self.CPUs, initializer=_mp_initializer, initargs=(image, self.classif_map, classifier)) as pool:
-        #     tiles_pred = pool.starmap(_predict_tile_mp,
-        #                               [(tilebounds, out_nodataVal, cmap_nodataVal)
-        #                                for tilebounds in get_array_tilebounds(array_shape=image.shape,
-        #                                                                       tile_shape=(1000, 1000))])
-        #
-        # for ((rS, rE), (cS, cE)), tile_pred in tiles_pred:
-        #     image_predicted[rS: rE + 1, cS: cE + 1, :] = tile_pred
-        #
-        # print(time.time() - t0)
-
         self.logger.info('Total prediction time: %s' % time.strftime("%H:%M:%S", time.gmtime(time.time()-t0)))
 
         ###############################
@@ -572,6 +617,9 @@ class RSImage_ClusterPredictor(object):
         # copy mask_nodata
         image_predicted.mask_nodata = image.mask_nodata
 
+        # append weights to predicted image
+        image_predicted.weights = weights
+
         # image_predicted.save(
         #     '/home/gfz-fe/scheffler/temp/SPECHOM_py/image_predicted_QRclust1_MinDist_noB9.bsq')
         # GeoArray(self.classif_map).save(
@@ -582,6 +630,18 @@ class RSImage_ClusterPredictor(object):
         cmap_valfractions = cmap_valcounts / self.classif_map.size
         self.classif_map_fractions = dict(zip(list(cmap_vals), list(cmap_valfractions)))
 
+        # log the pixel fraction where material-specific regressors were applied
+        frac = self.classif_map_fractions
+        if -1 in frac:
+            glob_regr_perc = frac[-1] * 100
+            nodata_perc = frac[cmap_nodataVal] * 100 if cmap_nodataVal in frac else 0
+            data_perc = 100 - nodata_perc
+            opt_regr_perc = 100 - glob_regr_perc - nodata_perc
+            self.logger.info(f"No-data fraction:\t{nodata_perc:.1f}%")
+            self.logger.info(f"Regressor fractions:\t"
+                             f"{opt_regr_perc / data_perc * 100:.1f}% material optimized; "
+                             f"{glob_regr_perc / data_perc * 100:.1f}% global regressor")
+
         return image_predicted
 
     def compute_prediction_errors(self, im_predicted, cluster_classifier, nodataVal=None, cmap_nodataVal=None):
@@ -636,25 +696,3 @@ class RSImage_ClusterPredictor(object):
         # GeoArray(errors).save('/home/gfz-fe/scheffler/temp/SPECHOM_py/errors_LRclust1_MinDist_noB9_clusterpred.bsq')
 
         return errors
-
-#
-# _global_image, _global_classif_map, _global_classifier = None, None, None
-#
-#
-# def _mp_initializer(image, classif_map, classifier):
-#     global _global_image, _global_classif_map, _global_classifier
-#     _global_image, _global_classif_map, _global_classifier = image, classif_map, classifier
-#
-#
-# def _predict_tile_mp(tilebounds, out_nodataVal, cmap_nodataVal):
-#     (rS, rE), (cS, cE) = tilebounds
-#     im_tile = _global_image[rS: rE + 1, cS: cE + 1, :]
-#     classif_map_tile = _global_classif_map[rS: rE + 1, cS: cE + 1]  # integer array
-#     classifier = _global_classifier
-#
-#     # predict!
-#     im_tile_pred = \
-#         classifier.predict(im_tile, classif_map_tile,
-#                            nodataVal=out_nodataVal, cmap_nodataVal=cmap_nodataVal).astype(_global_image.dtype)
-#
-#     return tilebounds, im_tile_pred
